require 'dotenv/load'
require 'octokit'
require 'faraday/http_cache'
require 'time'

module Rake
  module XPI
    class Config
      @@builds = :builds

      def octokit
        if @initialized.nil?
          raise "Github publishing not configured" unless @config.github && @config.github.user && @config.github.token
          token = ENV[@config.github.token]
          raise "Github publishing not configured" unless token

          Octokit.configure{|c|
            c.access_token = token
          }
          Octokit.middleware = Faraday::RackBuilder.new {|builder|
            builder.use Faraday::HttpCache, shared_cache: false
            builder.use Octokit::Response::RaiseError
            builder.adapter Faraday.default_adapter
          }
          Octokit.auto_paginate = true
          @initialized = true
          @repo = @config.github.user + '/' + @config.github.repo
        end
        return Octokit
      end

      def add_asset(tag, is_prerelease, asset_name, content, content_type)
        begin
          release = octokit.release_for_tag(@repo, tag.to_s)
        rescue Octokit::NotFound
          release = nil
        end
        if release
          raise "Release #{tag}: requested #{is_prerelease ? 'pre' : ''}release, found #{release.prerelease ? 'pre' : ''}release" if is_prerelease ^ release.prerelease
        else
          params = {
            name: tag.to_s,
            body: tag.to_s,
          }
          params[:target_commitish] = `git rev-list --max-parents=0 HEAD`.split(/\n/).collect{|sha| sha.strip}.first if tag.is_a?(Symbol)
          params[:prerelease] = true if is_prerelease
          release = octokit.create_release(@repo, tag.to_s, params)
        end

        # remove old version
        octokit.release_assets(release.url).each{|asset|
          next unless asset.name == asset_name
          octokit.delete_release_asset(asset.url)
        }

        octokit.upload_asset(release.url, content, name: asset_name, content_type: content_type)
      end

      def update_url
        return "https://github.com/#{@config.github.user}/#{@config.github.repo}/releases/download/update.rdf/update.rdf"
      end
      def xpi_url
        return "https://github.com/#{@config.github.user}/#{@config.github.repo}/releases/download/#{release_build? ? self.version : @@builds}/#{self.versioned_xpi}"
      end

      def publish
        raise "Github publishing not configured" unless @config.github && @config.github.user && @config.github.token
        token = ENV[@config.github.token]
        raise "Github publishing not configured" unless token

        update_rdf{|update| add_asset(:'update.rdf', false, 'update.rdf', update, 'application/rdf+xml') } if release_build?
        add_asset(release_build? ? self.release : @@builds, !release_build?, self.versioned_xpi, self.xpi, 'application/x-xpinstall')

        # scrub old builds after 7 days
        old = Time.now - (7*24*60*60)

        release = octokit.release_for_tag(@repo, @@builds.to_s)
        if release
          octokit.release_assets(release.url).each{|asset|
            next if asset.updated_at > old
            octokit.delete_release_asset(asset.url)
          }
        end
      end
    end
  end
end

